/*
 * timer.c - Event timers
 * 
 * Written 2002 by Werner Almesberger
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <signal.h>
#include <sys/time.h>

#include "timer.h"
#include "khelp.h"


#define __TRACE(s) \
  write(2,s,sizeof(s));
#define TRACE(s) __TRACE(s "\r\n");

#undef TRACE
#define TRACE(s)


typedef struct _timer {
    struct timeval due;
    void (*fn)(void *data);
    void *data;
    struct _timer *next;
} TIMER;


/* ----- Time arithmetic --------------------------------------------------- */


static double jiffie_f;
static struct timeval jiffie_tv;


static struct timeval jiffies2tv(unsigned long jiffies)
{
    struct timeval tv;
    double s;

    s = jiffies*jiffie_f;
    tv.tv_sec = floor(s);
    tv.tv_usec = floor((s-floor(s))*1e6);
    return tv;
}


unsigned long tv2jiffies(struct timeval tv)
{
    return floor(((double) tv.tv_sec+(double) tv.tv_usec/1e6)*(double) hz());
}


static struct timeval tv_add(struct timeval a,struct timeval b)
{
    a.tv_sec += b.tv_sec;
    a.tv_usec += b.tv_usec;
    while (a.tv_usec > 1000000) {
	a.tv_usec -= 1000000;
	a.tv_sec++;
    }
    return a;
}


static inline int tv_lt(const struct timeval a,const struct timeval b)
{
    if (a.tv_sec < b.tv_sec) return 1;
    if (a.tv_sec > b.tv_sec) return 0;
    return a.tv_usec < b.tv_usec;
}


void umlsim_setup_time(void)
{
    jiffie_f = 1.0/hz();
    jiffie_tv.tv_sec = 0;
    jiffie_tv.tv_usec = 1000000/hz();
    loops_per_jiffy = 1;
}


/* ----- Set simulation timers --------------------------------------------- */


static TIMER *timers = NULL;
struct timeval now = { 0,0 };


void set_timer(struct timeval tv,void (*fn)(void *data),void *data)
{
    TIMER *new,**walk;

    TRACE("set_timer");
    new = malloc(sizeof(TIMER));
    if (!new) abort();
    new->due = tv_add(now,tv);
    new->fn = fn;
    new->data = data;
    for (walk = &timers; *walk; walk = &(*walk)->next)
	if (!tv_lt((*walk)->due,new->due)) break;
    new->next = *walk;
    *walk = new;
}


void wake_kernel(unsigned delta_jiffies)
{
    TRACE("wake_kernel");
    set_timer(jiffies2tv(delta_jiffies),NULL,NULL);
}


/* ----- Next simulation event --------------------------------------------- */


int use_sigalrm = 0; /* tap.c signals back which signal UML expects */
int dont_tick = 0; /* signal tap.c whether to do_timer on timer interrupt */

static struct timeval last_tick = { 0,0 };
static int timer_pending = 0;


static void do_tick(int tick)
{
    if (tick) {
	timer_pending = 0;
	last_tick = now;
    }
    else {
	__TRACE("do_tick(0)");
	dont_tick = 1;
    }
    raise(use_sigalrm ? SIGALRM : SIGVTALRM);
}


static void have_timers(void)
{
    TIMER *curr;
    struct timeval next_tick;

    next_tick = tv_add(last_tick,jiffie_tv);
    while (tv_lt(next_tick,timers->due)) {
	now = next_tick;
	if (got_work_with_tick() || timer_pending) {
	    do_tick(1);
	    return;
	}
	jiffies++;
	last_tick = now;
	next_tick = tv_add(now,jiffie_tv);
    }
    curr = timers;
    timers = timers->next;
    if (!curr->fn) timer_pending = 1;
    else {
	now = curr->due;
	curr->fn(curr->data);
    }
    free(curr);
}


static void no_timers(void)
{
    struct timeval next_tick;

    next_tick = tv_add(last_tick,jiffie_tv);
    if (got_work_with_tick() || timer_pending) {
	now = next_tick;
	do_tick(1);
	return;
    }
    __TRACE("THE SYSTEM SEEMS IDLE, WHICH IS IMPOSSIBLE.\r\n");
}


void wait_kernel(void)
{
    TRACE("wait_kernel");
    if (other_runnable()) return;
    if (got_work_without_tick()) {
	do_tick(0);
	return;
    }
    if (timers) have_timers();
    else no_timers();
}
